"""
Provides functions for undo-ing decay correction and recalculating it.

"""
from typing import Optional
import math

import ants
import numpy as np
import pandas as pd

from ..utils import image_io
from ..utils.scan_timing import ScanTimingInfo
from ..io.image import ImageLoader
from ..utils.image_io import safe_copy_meta
from petpal.utils.constants import HALF_LIVES
from ..meta.auto_cli import auto_cli

def undo_decay_correction(input_image_path: str,
                          output_image_path: str,
                          metadata_dict: dict = None) -> ants.ANTsImage:
    """Uses decay factors from the metadata for an image to remove decay correction for each frame.

    This function expects to find decay factors in the .json sidecar file, or the metadata_dict, if given. If there are
    no decay factors (either under the key 'DecayFactor' or the BIDS-required 'DecayCorrectionFactor') listed, it may
    result in unexpected behavior. In addition to returning an ANTsImage containing the "decay uncorrected" data, the
    function writes an image to output_image_path, unless it is passed as 'None'.

    Args:
        input_image_path (str): Path to input (.nii.gz or .nii) image. A .json sidecar file should exist in the same
             directory as the input image.
        output_image_path (str): Path to output (.nii.gz or .nii) output image. If None, no image will be written.
        metadata_dict (dict, optional): Metadata dictionary to use instead of corresponding .json sidecar. If not
            specified (default behavior), function will try to use sidecar .json in the same directory as
            input_image_path.

    Returns:
        ants.ANTsImage: ANTsImage with decay correction reversed."""

    decay_corrected_image = ants.image_read(filename=input_image_path)

    if metadata_dict is not None:
        json_data = metadata_dict
    else:
        json_data = image_io.load_metadata_for_nifti_with_same_filename(image_path=input_image_path)

    frame_info = ScanTimingInfo.from_nifti(image_path=input_image_path)
    decay_factors = frame_info.decay

    uncorrected_image_numpy = decay_corrected_image.numpy()

    for frame_num, decay_factor in enumerate(decay_factors):
        uncorrected_image_numpy[..., frame_num] /= decay_factor

    uncorrected_image = ants.from_numpy_like(data=uncorrected_image_numpy,
                                             image=decay_corrected_image)

    if output_image_path is not None:
        ants.image_write(image=uncorrected_image,
                         filename=output_image_path)

        json_data['DecayFactor'] = list(np.ones_like(decay_factors))
        json_data['ImageDecayCorrected'] = "false"
        output_json_path = image_io.gen_meta_data_filepath_for_nifti(nifty_path=output_image_path)
        image_io.write_dict_to_json(meta_data_dict=json_data,
                                    out_path=output_json_path)

    return uncorrected_image


def decay_correct(input_image_path: str,
                  output_image_path: str) -> ants.ANTsImage:
    r"""Recalculate decay_correction for nifti image based on frame reference times.

    This function will compute frame reference times based on frame time starts and frame durations (both of which
    are required by BIDS. These reference times are used in the following equation to determine the decay factor for
    each frame. For more information, refer to Turku Pet Centre's materials at
    https://www.turkupetcentre.net/petanalysis/decay.html

    .. math::
        decay\_factor = \exp(\lambda*t)

    where :math:`\lambda=\log(2)/T_{1/2}` is the decay constant of the radio isotope and depends on its half-life and
    `t` is the frame's reference time with respect to TimeZero (ideally, injection time).

    Note: BIDS 'DecayCorrectionTime' is set to 0 (seconds w.r.t. TimeZero) for the image. If this assumption doesn't
        hold, be wary of downstream effects.

    Args:
        input_image_path (str): Path to input (.nii.gz or .nii) image. A .json sidecar file should exist in the same
             directory as the input image.
        output_image_path (str): Path to output (.nii.gz or .nii) output image.

    Returns:
        ants.ANTsImage: Decay-Corrected Image

    """
    half_life = image_io.get_half_life_from_nifti(image_path=input_image_path)

    json_data = image_io.load_metadata_for_nifti_with_same_filename(image_path=input_image_path)
    uncorrected_image = ants.image_read(filename=input_image_path)

    frame_info = ScanTimingInfo.from_nifti(image_path=input_image_path)
    frame_reference_times = np.asarray(frame_info.start + frame_info.duration / 2.0, float).tolist()

    original_decay_factors = frame_info.decay
    if np.any(original_decay_factors != 1):
        raise ValueError(f'Decay Factors other than 1 found in metadata for {input_image_path}. This likely means the '
                         f'image has not had its previous decay correction undone. Try running undo_decay_correction '
                         f'before running this function to avoid decay correcting an image more than once.')

    corrected_data = uncorrected_image.numpy()
    new_decay_factors = []
    for frame_num, frame_reference_time in enumerate(frame_reference_times):
        new_decay_factor = math.exp(((math.log(2) / half_life) * frame_reference_time))
        corrected_data[..., frame_num] *= new_decay_factor
        new_decay_factors.append(new_decay_factor)

    corrected_image = ants.from_numpy_like(data=corrected_data,
                                           image=uncorrected_image)

    if output_image_path is not None:
        ants.image_write(image=corrected_image,
                         filename=output_image_path)
        output_json_path = image_io.gen_meta_data_filepath_for_nifti(nifty_path=output_image_path)
        json_data['DecayFactor'] = new_decay_factors
        json_data['ImageDecayCorrected'] = "true"
        json_data['ImageDecayCorrectionTime'] = 0
        json_data['FrameReferenceTime'] = frame_reference_times
        image_io.write_dict_to_json(meta_data_dict=json_data,
                                    out_path=output_json_path)

    return corrected_image


def calculate_frame_decay_factor(frame_reference_time: np.ndarray,
                                 half_life: float) -> np.ndarray:
    """Calculate decay correction factors for a scan given the frame reference time and half life.
    
    Important: 
        The frame reference time should be the time at which average activity occurs,
        not simply the midpoint. See
        :meth:`~petpal.utils.scan_timing.calculate_frame_reference_time` for more info.

    Args: 
        frame_reference_time (np.ndarray): Time at which the average activity occurs for the frame.
        half_life (float): Radionuclide half life.

    Returns: 
        np.ndarray: Decay Correction Factors for each frame in the scan.
    """
    decay_constant = np.log(2)/half_life
    frame_decay_factor = np.exp((decay_constant)*frame_reference_time)
    return frame_decay_factor


def scale_frames(input_img: ants.ANTsImage, scalar_arr: np.ndarray[float]):
    nframes = input_img.shape[-1]
    nscalar = len(scalar_arr)
    if nframes!=nscalar:
        raise ValueError(f"Length of correction factors ({nscalar}) does not "
                         f"match number of frames in dynamic PET ({nframes}).")
    modified_arr = ants.image_clone(input_img)
    for frame, scalar in enumerate(scalar_arr):
        modified_arr[:,:,:,frame] *= scalar
    return modified_arr


class DecayCorrect:
    """Decay correct or uncorrect each frame in a dynamic PET scan.
    
    :ivar image_loader: The image loader to use.
    :ivar modified_pet_img: The corrected PET image."""
    def __init__(self,
                 image_loader: Optional[ImageLoader] = None):
        self.image_loader = image_loader or ImageLoader()
        self.modified_pet_img: ants.ANTsImage = None

    def apply_factor(self,
                     input_image_path: str,
                     correction_factor: np.ndarray[float]):
        """Apply fix factor to image and write output.

        Args:
            input_image_path (str): Path to dynamic PET image.
            correction_factor_path (str): Path to file with correction factors, one per frame. 
                File must have exactly one column, with a header followed by one scalar per
                line."""

        input_img = self.image_loader.load(filename=input_image_path)
        self.modified_pet_img = scale_frames(input_img=input_img, scalar_arr=correction_factor)
    
    def save_modified_pet(self, input_image_path: str, output_image_path: str):
        """Save the modified PET image

        Args:
            input_image_path (str): Path to dynamic PET image.
            output_image_path (str): Path to where corrected image is saved.
        """
        ants.image_write(self.modified_pet_img, output_image_path)
        safe_copy_meta(input_image_path,out_image_path=output_image_path)

    def decay_correct_factor_from_file(self,
                                       input_image_path: str,
                                       output_image_path: str,
                                       correction_factor_path: str):
        """Apply a set of correction factors to each frame in a PET image.
        
        Provide path to dynamic PET image, path to where corrected image is saved, and path to a
        file containing factors for each frame. File containing correction factors must have
        exactly one column, with a header followed by one scalar per line to apply to the
        corresponding frame in the PET. Reads image and correction factors, applies them, and saves
        result.
        
        Args:
            input_image_path (str): Path to dynamic PET image.
            output_image_path (str): Path to where corrected image is saved.
            correction_factor_path (str): Path to file with correction factors, one per frame. 
                File must have exactly one column, with a header followed by one scalar per
                line.
        """
        correction_factor = pd.read_csv(correction_factor_path,index_col=False).iloc[:,0]
        self.apply_factor(input_image_path=input_image_path,
                          correction_factor=correction_factor)
        self.save_modified_pet(input_image_path=input_image_path,
                               output_image_path=output_image_path)


class DecayFix(DecayCorrect):
    """Special case for decay correction where the dynamic PET image was scaled with the wrong
    isotope but JSON metadata retains accurate decay correction factors."""
    def fix_factor(self,
                   input_image_path: str,
                   isotope_to_remove: str) -> np.ndarray[float]:
        """Get ratio of the expected decay correction factor to the factor that was incorrectly
        applied to PET image.
        
        Args:
            input_image_path (str): Path to dynamic PET image.
            isotope_to_remove (str): Name of the isotope that was incorrectly used to scale dynamic
                PET image, such as 'o15' or 'c11'.
        
        Returns:
            fix_factor (np.ndarray[float]): Scalar array with the correction factor to apply to
                each frame in dynamic PET.
        """
        scan_timing = ScanTimingInfo.from_nifti(input_image_path)
        decay_expected = scan_timing.decay
        decay_applied = calculate_frame_decay_factor(scan_timing.center,
                                                     HALF_LIVES[isotope_to_remove])
        return decay_expected/decay_applied

    def __call__(self,                 
                input_image_path: str,
                output_image_path: str,
                isotope_to_remove: str):
        """Calculate correction factor based on the accurate metadata and name of the isotope
        incorrectly used to scale image data.

        Args:
            input_image_path (str): Path to dynamic PET image.
            output_image_path (str): Path to where corrected image is saved.
            isotope_to_remove (str): Name of the isotope that was incorrectly used to scale dynamic
                PET image, such as 'o15' or 'c11'.
        """
        correction_factor = self.fix_factor(input_image_path=input_image_path,
                                            isotope_to_remove=isotope_to_remove)
        self.apply_factor(input_image_path=input_image_path,
                          correction_factor=correction_factor)
        self.save_modified_pet(input_image_path=input_image_path,
                               output_image_path=output_image_path)

def main():
    auto_cli(petpal_class=DecayFix)

if __name__=='__main__':
    main()



"""Wrappers for generating CLI functions from PETPAL callable objects"""
import argparse
import logging
import inspect
import sys
import copy
from pydoc import locate


def camel_to_kebab_case(name: str) -> str:
    """Convert a string in camel case to kebab case.
    
    Camel case is used for class objects in Python, e.g. MyClass.
    Kebab case is used for programs, such as my-prog.
    This function converts a string in camel case to kebab case, for the purpose of assigning
    CLI program names to callable class objects.
    
    Args:
        name (str): The camel case phrase to be converted to kebab case.
    
    Returns:
        kebab_name (str): Input phrase converted to kebab case.
    """
    if len(name)==1:
        return name.lower()

    kebab_name = name[0].lower()
    for char in name[1:]:
        if char.isupper():
            kebab_name += f'-{char.lower()}'
        else:
            kebab_name += char
    return kebab_name


def args_kwargs_to_dictionary(args: argparse.Namespace) -> dict:
    """Convert Namespace to dictionary and reassign keyword arguments to dictionary values.
    
    Args:
        args (argparse.Namespace): The args resulting from parsing command line arguments.
    
    Returns:
        call_eval (dict): Dictionary with arguments and keyword arguments from inspecting the class
            __call__ function.
    """
    arg_vals = vars(args)
    call_eval = arg_vals.copy()

    for arg in arg_vals:
        value = arg_vals[arg]
        if isinstance(value, dict):
            for kwarg in value:
                call_eval[kwarg] = value[kwarg]
            call_eval.pop(arg)

    return call_eval


class ParseKwargs(argparse.Action):
    """Action to parse keyword arguments."""
    SUPPORTED_KWARG_TYPES = [str, float, int, bool]

    def __call__(self, parser, namespace, values, option_string=None):
        """Creates a dictionary within the args namespace holding keyword arguments.
        
        Kwargs are used like so: --kwargs int:frame=4
        Each kwarg must specify type before a colon, followed by the argument name, an equals sign
        to delimit the argument value. Kwargs are currently limited to a small number of standard
        types: str, float, int, bool."""
        setattr(namespace, self.dest, {})
        for value in values:
            kwarg_type, kwarg_pair = value.split(':')
            kwarg_locator = locate(kwarg_type)
            kwarg_name, kwarg_value = kwarg_pair.split('=')

            if kwarg_locator not in self.SUPPORTED_KWARG_TYPES:
                raise TypeError(f"CLI Kwargs only supports types: {self.SUPPORTED_KWARG_TYPES}."
                                f"Got {kwarg_locator}.")

            getattr(namespace, self.dest)[kwarg_name] = kwarg_locator(kwarg_value)


class PetpalLogging:
    """Tool for logging CLI commands.
    
    Records PETPAL CLI commands in a log file. Creates a new logfile at the provided path, if one 
    does not already exist. If it points to an existing file, any logging is added to the bottom of
    the file. Logs command line arguments only."""
    def __init__(self, logfile: str=None):
        if logfile is not None:
            logging.basicConfig(filename=logfile, level=logging.INFO, format='')
            self.logger = logging.getLogger('')
            self.cli_args = copy.deepcopy(sys.argv)
            self.log_cli_args()

    def log_cli_args(self):
        """Log the command line that was used to run PETPAL.
        
        Records the literal PETPAL command line tool, and any interpolated variables as they are
        interpreted by Python.
        """
        self.logger.info(' '.join(self.cli_args))


def auto_cli(petpal_class: object):
    """Generate a command line interface for a PETPAL function
    
    Args:
        petpal_class (object): Class defined in PETPAL that can be instantiated without specifying
            __init__ arguments. Must contain function __call__ with a docstring.

    Important:
        The provided class must be designed such that any __init__ arguments can be set with
        defaults, and those defaults should only be changed for testing and development purposes.
        Otherwise, it would be impossible to tell which CLI arguments should be assigned to the
        __init__ function and which should be assigned to __call__.

        Additionally, any __call__ arguments and keyword arguments must be able to be specified from
        the command line, such as strings, integers, or floats.

        Keyword arguments are only intended to be used with the CLI when it would be excessive and
        cumbersome to implement all possible arguments of a function. Any expected user inputs
        should be specified, documented, and type hinted in the __call__ args, reserve kwargs only
        for providing user flexibility where necessary, as it carries the risk of unintended
        behavior.

    Example:
        
        .. code-block:: python

            import numpy as np
            from petpal.meta.auto_cli import auto_cli
            import external_func

            class MyClass:
                mri_img: ants.ANTsImage
                pet_img: ants.ANTsImage

                def my_func(self, **kwargs):
                    return external_func(self.mri_img, self.pet_img, **kwargs)

                def __call__(self, mri_img_path, pet_img_path, out_img_path, **kwargs):
                    self.mri_img = ants.image_read(mri_img_path)
                    self.pet_img = ants.image_read(pet_img_path)
                    output_img = my_func(**kwargs)
                    ants.image_write(output_img, output_img)
            
            def main():
                # Creates CLI for class my_class
                # __call__ args are interpreted as required arguments
                # **kwargs is interpreted as optional keyword arguments
                # usage: python my_class.py my-class --mri-img-path [MRI_IMG_PATH] \\
                #          --pet-img-path [PET_IMG_PATH] --kwargs [kwarg1=val1 kwarg2=val2 ...]
                # Add file path to pyproject.toml to generate a command shortcut

                auto_cli(petpal_class=my_class)

            if __name__=='__main__':
                main()
    """
    parser = argparse.ArgumentParser(prog=petpal_class.__name__,
                                     description=petpal_class.__call__.__doc__,
                                     formatter_class=argparse.RawTextHelpFormatter)

    call_signature = inspect.signature(petpal_class.__call__)
    call_parameters = call_signature.parameters.items()
    for _, call_parameter in call_parameters:
        arg_name = str(call_parameter)
        if arg_name=='self':
            continue
        arg_and_type = arg_name.split(': ')
        if len(arg_and_type)==2:
            arg_name = f'--{arg_and_type[0]}'.replace('_','-')
            arg_type = locate(arg_and_type[1])
            parser.add_argument(arg_name,type=arg_type,required=True)
        elif arg_and_type[0].startswith('**'):
            kwarg_name = arg_and_type[0].replace('**','--').replace('_','-')
            parser.add_argument(kwarg_name, nargs='*', action=ParseKwargs, required=False)
    parser.add_argument('--logfile',type=str,default=None,required=False)

    args = parser.parse_args()
    arg_vals = args_kwargs_to_dictionary(args=args)

    logger = PetpalLogging(logfile=arg_vals['logfile'])
    arg_vals.pop('logfile')

    init_class = petpal_class()
    init_class(**arg_vals)
